/* Copyright 2016 Kristofer Björnson
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** @file HoppingAmplitude.cpp
 *
 *  @author Kristofer Björnson
 */

#include "HoppingAmplitude.h"
#include "Streams.h"

#include <sstream>

#include "json.hpp"

using namespace std;
using namespace nlohmann;

namespace TBTK{

HoppingAmplitude::HoppingAmplitude(
	complex<double> amplitude,
	Index toIndex,
	Index fromIndex
) :
	fromIndex(fromIndex),
	toIndex(toIndex)
{
	this->amplitude = amplitude;
	this->amplitudeCallback = NULL;
};

HoppingAmplitude::HoppingAmplitude(
	complex<double> (*amplitudeCallback)(Index, Index),
	Index toIndex,
	Index fromIndex
) :
	fromIndex(fromIndex),
	toIndex(toIndex)
{
	this->amplitudeCallback = amplitudeCallback;
};

HoppingAmplitude::HoppingAmplitude(
	const HoppingAmplitude &ha
) :
	fromIndex(ha.fromIndex),
	toIndex(ha.toIndex)
{
	amplitude = ha.amplitude;
	this->amplitudeCallback = ha.amplitudeCallback;
}

HoppingAmplitude::HoppingAmplitude(
	const string &serialization,
	Serializeable::Mode mode
){
	TBTKAssert(
		Serializeable::validate(
			serialization,
			"HoppingAmplitude",
			mode
		),
		"HoppingAmplitude::HoppingAmplitude()",
		"Unable to parse string as HoppingAmplitude '"
		<< serialization << "'.",
		""
	);

	switch(mode){
	case Serializeable::Mode::Debug:
	{
		string content = Serializeable::getContent(serialization, mode);

		vector<string> elements = Serializeable::split(
			content,
			Serializeable::Mode::Debug
		);

		amplitudeCallback = nullptr;

		stringstream ss;
		ss.str(elements.at(0));
		ss >> amplitude;
		toIndex = Index(elements.at(1), mode);
		fromIndex = Index(elements.at(2), mode);
		break;
	}
	case Serializeable::Mode::JSON:
	{
		try{
			amplitudeCallback = nullptr;

			json j = json::parse(serialization);
			Serializeable::deserialize(j["amplitude"].get<string>(), &amplitude, mode);
			toIndex = Index(j["toIndex"].dump(), mode);
			fromIndex = Index(j["fromIndex"].dump(), mode);
		}
		catch(json::exception e){
			TBTKExit(
				"HoppingAmplitude::HoppingAmplitude()",
				"Unable to parse string as HoppingAmplitude '"
				<< serialization << "'.",
				""
			);
		}

		break;
	}
	default:
		TBTKExit(
			"HoppingAmplitude::HoppingAmplitude()",
			"Only Serializeable::Mode::Debug is supported yet.",
			""
		);
	}
}

HoppingAmplitude HoppingAmplitude::getHermitianConjugate() const{
	if(amplitudeCallback)
		return HoppingAmplitude(amplitudeCallback, fromIndex, toIndex);
	else
		return HoppingAmplitude(conj(amplitude), fromIndex, toIndex);
}

void HoppingAmplitude::print() const{
	Streams::out << "From index:\t";
	for(unsigned int n = 0; n < fromIndex.getSize(); n++){
		Streams::out << fromIndex.at(n) << " ";
	}
	Streams::out << "\n";
	Streams::out << "To index:\t";
	for(unsigned int n = 0; n < toIndex.getSize(); n++){
		Streams::out << toIndex.at(n) << " ";
	}
	Streams::out << "\n";
	Streams::out << "Amplitude:\t" << getAmplitude() << "\n";
}

string HoppingAmplitude::serialize(Serializeable::Mode mode) const{
	TBTKAssert(
		amplitudeCallback == nullptr,
		"HoppingAmplitude::serialize()",
		"Unable to serialize HoppingAmplitude that uses callback"
		<< " value.",
		""
	);

	switch(mode){
	case Serializeable::Mode::Debug:
	{
		stringstream ss;
		ss << "HoppingAmplitude(";
		ss << Serializeable::serialize(amplitude, mode) << ",";
		ss << toIndex.serialize(mode) << "," << fromIndex.serialize(mode);
		ss << ")";

		return ss.str();
	}
	case Serializeable::Mode::JSON:
	{
		json j;
		j["id"] = "HoppingAmplitude";
		j["amplitude"] = Serializeable::serialize(amplitude, mode);
		j["toIndex"] = json::parse(toIndex.serialize(mode));
		j["fromIndex"] = json::parse(fromIndex.serialize(mode));

		return j.dump();

/*		stringstream ss;
		ss << "{";
		ss << "id:'HoppingAmplitude'";
		ss << "," << "amplitude:";
		ss << amplitude;
		ss << "," << "to:" << toIndex.serialize(mode);
		ss << "," << "from:" << fromIndex.serialize(mode);
		ss << "}";

		return ss.str();*/
	}
	default:
		TBTKExit(
			"HoppingAmplitude::serialize()",
			"Only Serializeable::Mode::Debug is supported yet.",
			""
		);
	}
}

};	//End of namespace TBTK
