/* Copyright 2016 Kristofer Björnson
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** @file DOS.cpp
 *
 *  @author Kristofer Björnson
 */

#include "DOS.h"
#include "Streams.h"

#include "json.hpp"

using namespace std;
using namespace nlohmann;

namespace TBTK{
namespace Property{

DOS::DOS(
	double lowerBound,
	double upperBound,
	int resolution
) :
	AbstractProperty(resolution)
{
	this->lowerBound = lowerBound;
	this->upperBound = upperBound;
	this->resolution = resolution;
}

DOS::DOS(
	double lowerBound,
	double upperBound,
	int resolution,
	const double *data
) :
	AbstractProperty(resolution, data)
{
	this->lowerBound = lowerBound;
	this->upperBound = upperBound;
	this->resolution = resolution;
}

DOS::DOS(
	const DOS &dos
) :
	AbstractProperty(dos)
{
	lowerBound = dos.lowerBound;
	upperBound = dos.upperBound;
	resolution = dos.resolution;
}

DOS::DOS(
	DOS &&dos
) :
	AbstractProperty(std::move(dos))
{
	lowerBound = dos.lowerBound;
	upperBound = dos.upperBound;
	resolution = dos.resolution;
}

DOS::DOS(
	const string &serialization, Mode mode
) :
	AbstractProperty(
		Serializeable::extract(
			serialization,
			mode,
			"abstractProperty"
		),
		mode
	)
{
	TBTKAssert(
		validate(serialization, "DOS", mode),
		"DOS::DOS()",
		"Unable to parse string as DOS '" << serialization << "'.",
		""
	);

	switch(mode){
	case Mode::JSON:
		try{
			json j = json::parse(serialization);
			lowerBound = j.at("lowerBound").get<double>();
			upperBound = j.at("upperBound").get<double>();
			resolution = j.at("resolution").get<int>();
		}
		catch(json::exception e){
			TBTKExit(
				"DOS::DOS()",
				"Unable to parse string as DOS '"
				<< serialization << "'.",
				""
			);
		}

		break;
	default:
		TBTKExit(
			"DOS::DOS()",
			"Only Serializeable::Mode::JSON is supported yet.",
			""
		);
	}
}

DOS::~DOS(){
}

DOS& DOS::operator=(const DOS &rhs){
	if(this != &rhs){
		AbstractProperty::operator=(rhs);
		lowerBound = rhs.lowerBound;
		upperBound = rhs.upperBound;
		resolution = rhs.resolution;
	}

	return *this;
}

DOS& DOS::operator=(DOS &&rhs){
	if(this != &rhs){
		AbstractProperty::operator=(std::move(rhs));
		lowerBound = rhs.lowerBound;
		upperBound = rhs.upperBound;
		resolution = rhs.resolution;
	}

	return *this;
}

string DOS::serialize(Mode mode) const{
	switch(mode){
	case Mode::JSON:
	{
		json j;
		j["id"] = "DOS";
		j["lowerBound"] = lowerBound;
		j["upperBound"] = upperBound;
		j["resolution"] = resolution;
		j["abstractProperty"] = json::parse(
			AbstractProperty::serialize(mode)
		);

		return j.dump();
	}
	default:
		TBTKExit(
			"DOS::serialize()",
			"Only Serializeable::Mode::JSON is supported yet.",
			""
		);
	}
}

};	//End of namespace Property
};	//End of namespace TBTK
